/*
Class:   AutoFillSeek
Author:  Rich Assaf (toytracker@attbi.com)
History: 
   2.0  January 1998
      AutoFill.ZIP (VdB 5.x) posted in CompuServe dBase
   3.0  April 15, 2002
      Initial public release for dB2K
   3.1  April 17, 2002
   	Modified DisplayEF to handle a trailing blank
      space, which will advance to the next record.
Purpose:
   Originally inspired by the QuickFill feature in 
   Quicken, this control speeds data entry by 
   displaying matching data in an entryfield based
   on the first few typed characters.

Credits:
   I'd been hoping that someone would recreate this 
   functionality for dB2K, and some posts in the NG
   did get me started in this direction.  Then I found
   Dan Howard's excellent GridCombo in dBulletin, and 
   much of the difficult work had been done.... And of
   course, without Ken Chan's Seeker and Romain's help
   in the old CompuServe forum, none of this would have
   been possible for me at all.  A big "THANK YOU"!
   See the "Notes" section below for an outline of 
   where this code diverges from Dan's.

Requirements:
   down.bmp  (Down Arrow bitmap - Dan Howard's)
   font.cc (Dan Howard's)
   dhform.cfm (Dan Howard's)
   	(or at least functions Freeze, Unfreeze and 
      Repaint, and call the control's init(). Dan's 
      DHForm does all this (and more!), see especially
      his Iterate() function )

Usage:
   1. set procedure to autofillseek.cc additive
   2. Drop one on the form, datalink the grid and define
   at least one column.  Although the grid may not appear
   in the form designer, you can select it by clicking 
   just below the container, which will highlight the 
   outline of the grid.
   3. Optionally set the datalink for the DATAFIELD 
   object.  In design mode, the DATAFIELD object has a 
   cyan background (and default value of "Datalink Me!").
   By default, it fills the first half of the container.
   Everything is resized at run time to look like a 
   standard combobox.
   4. Optionally, define a NoMatch codeblock or function
   pointer, e.g. FORM.AUTOFILLSEEK1.NoMatch = CLASS::DoIt
   This code will be executed if appropriate (see the
   demo form AFDEMO for an example of this).
   5. Set the font type and size in the SEEKFIELD, and 
   the information will be carried over to the other
   controls (unless CarryFontInformation is set false).
   Thanks again to Dan's Font.CC..


Properties:

   DropDownHeight - (default 220 pixels, about 10 items) 
   	Specifies the dropdown height of the Grid in
      PIXEL metrics.

	IsGridOpen -  (default is false)		
   	Whether or not the Grid is currently open

   CarryFontInformation -  (default is true)
   	If set to true , the font information specified
      in the entryfield will be carried over into 
      the grid columns

   RefreshAlways - (default is true)
   	If set to True, calls form.rowset.refreshControls() 
      when the value changes.

   lMatch - (default is true)
   	Property to track whether or not a match has been
      located after each character is processed.  Use to
      optionally execute some code (e.g. open an "Add" 
      form) if lMatch is false when container loses focus.


Notes/Changes from GridCombo:

   - Added an additional entryfield for datalinking. 
   	For whatever reason, datalinking the seeker field
		breaks the seek function, so work around this by
      maintaining a second "DATAFIELD" for displaying
      the datalinked info.  The "SEEKFIELD" takes over
      when appropriate....

   - Added a DisplayEF method to show the partially
   	entered information with the remaining AutoFill 
		information highlighted, and cursor placed correctly.

   - Added code in Init() to correctly place the
   	dropdown grid no matter how deeply into
      the form hierarchy the container is placed
      (i.e. object in a container on a notebook etc).

   - Added a property lMatch, to check if a match was
   	detected.  If not, execute the user-defined
      NoMatch code if it has been defined.

   - Changed behavior of Enter key - if dropdown grid
   	is closed, it Tabs to next control. If the
      grid is displayed, selected value becomes the 
      entryfield value and the grid is closed.

   - Clicking anywhere in the entryfield does NOT drop
   	down the grid as in GridCombo; must specifically 
      click the DownArrow (or use keyboard DnArrow)

   - Minor modifications of MoveRecord function, to 
   	test for endofSet and prevent using the arrow
      keys to move to a blank value.

   - Still requires PIXELS (the default if using 
   	DHForm.cfm)

   - Still doesn't handle LookupSQL type of combobox...

   - Creates an _app.GridCombo as a pointer to "this".
   	Needed for the arrow keys.  It gets nullified 
      when the entryfield loses focus.


Known Bug:

	- Clicking away from the control when the grid is
   	open does not close the grid.  See AF_SEEKER
      onLostFocus for details.  Suggestions welcome..

The author accepts no responsibility for the use or 
consequences of the use of this software.

copyright 2002 - Richard Assaf
*/

class AutoFillSeek(parentObj) of CONTAINER(parentObj) custom
   with (this)
      width = 120
      height = 22
      metric = 6	// Pixels
      borderStyle = 4	// Single
   endwith

   this.DATAFIELD = new AF_DATA(this)
   with (this.DATAFIELD)
      width = 60
      fontSize = 9
      value = "DataLink Me!"
      colornormal = "Black/Cyan"
   endwith

   this.SEEKFIELD = new AF_SEEKER(this)
   with (this.SEEKFIELD)
      height = 20
      left = 60
      width = 60
      value = ""
   endwith

   this.GRID = new AF_GRID(this)
   with (this.GRID)
      bgColor = "white"
      cellHeight = 20
      pageno = 0
      height = 8
      left = 0
      top = 21
      width = 120
   endwith


   this.IsGridOpen = false
   this.onChange = NULL
   this.DropDownHeight = 220 // default to about 10 items
   this.CarryFontInformation = true
   this.RefreshAlways = true
   this.lMatch = true  // For the container


   Function Init
      // Make sure minimum requirements are met
      if type("this.Grid.DataLink") == "U"
      	msgbox("AutoFillSeek requires a datalink "+chr(13)+ ;
         		"for the grid", "Critical Error",16)
         RETURN
      endif
      if type("this.Grid.Columns[1].DataLink") == "U"
      	msgbox("AutoFillSeek requires a datalink "+chr(13)+ ;
         		"for one column for the grid", "Critical Error",16)
         RETURN
      endif

      // Hide the container's border at run-time.
      this.BorderStyle = 3 // None

      // Create the drop button
      this.DropBUTTON = new PUSHBUTTON(this)
      with (this.DropButton)
         onClick = class::DropButton_ONCLICK
         borderStyle = 2
         metric = 6
         height = this.height - 4
         width = 20
         left = this.Width - width //- 1
         top = 2 //this.SEEKFIELD.Top + 1
         text = ""
         speedBar = true
         upBitmap = "filename Down.bmp"
         fontSize = 9
         pageno = 0
         value = false
      endwith

      // Reposition the SEEKFIELD to size of  container
      this.SEEKFIELD.Left   := 1 // Left edge of container
      this.SEEKFIELD.Width  := this.Width - this.dropbutton.width+1
      this.SEEKFIELD.Height := this.Height-2 

      // Same for the DATAFIELD too ...
      this.DATAFIELD.Left   := 1 // Left edge of container
      this.DATAFIELD.Width  := this.Width - this.dropbutton.width+1
      this.DATAFIELD.Height := this.Height-2

      // Assign the rowset to the seeker
      if type("this.Grid.DataLink") # "U"
         // rowset to perform seeker operations
         this.SEEKFIELD.rowset := this.Grid.DataLink
      endif

      // Prototype the WinAPI to position grid
      if type('WinAPISetParent') # 'FP'
         extern chandle WinAPISetParent(chandle, chandle) user32 from 'SetParent'
      endif

      // Set the correct grid Top and Left position
      oObj = this ; nLeft = 0; nTop = 0
      do // while true
      	nTop += oObj.Top
         nLeft += oObj.Left
         oObj = oObj.parent
      until oObj = form

      this.Grid.Top    := nTop + this.Height
      this.Grid.Left   := nLeft
      this.Grid.Width  := this.Width
      this.Grid.Height := this.DropDownHeight
      this.Grid.CellHeight := this.SEEKFIELD.Height - 1

      /* If CarryFontInformation is true, then carry all 
      of the font information from the SEEKFIELD and 
      assign it to the GRID columns and DATAFIELD
      */
      if this.CarryFontInformation
         local f,j
         set procedure to font.cc additive
         f = New Font()
         f.DefineFrom(this.SEEKFIELD)
         for j = 1 to this.Grid.Columns.Size
            f.SetObjectFont(this.Grid.Columns["Column"+j].EditorControl)
         next j
         f.SetObjectFont(this.DATAFIELD)
      endif
   return


   function DropButton_onClick
      if this.Parent.DATAFIELD.visible
      	this.Parent.DATAFIELD.onGotFocus()
      endif

      this.Parent.ToggleGrid()
   return

   Function ToggleGrid
      // optional parameter to force grid open or closed
      // NB: problem with handling a logical parameter,
      //  type() is defined with value of false even if
      //  no parameter is passed.  Use chars as workaround. 
      parameters cToggle // receive as "T" or "F"
   	if type("cToggle")=="C" // parameter passed
         cToggle := upper(cToggle)
      	if (iif(cToggle="T",.t.,.f.) = this.IsGridOpen )
            cToggle = null
      		RETURN // ignore if grid is in desired state
         endif
      endif

      if not this.IsGridOpen
         this.Grid.Visible = false // make sure I'm hidden
         local bm
         bm = this.Grid.DataLink.BookMark()
         // Kill any OnSelChange Event defined because it will
         // fire here. We don't want that since the user of this
         // component will probably use onSelChange for his/her
         // own purposes. This blop of code is really just to 
         // scroll the grid properly.
         local oldOnSelChange
         oldOnSelChange = this.Grid.OnSelChange
         this.Grid.OnSelChange := NULL
         this.Grid.DataLink.First()
         this.Grid.DataLink.Goto(bm)
         this.Grid.OnSelChange := oldOnSelChange
         this.Grid.Visible := true
         WinApiSetParent(this.Grid.hWnd,form.hWnd)
         this.SEEKFIELD.SetFocus()  
      else
         this.Grid.Visible := false
      endif

      this.IsGridOpen := not this.IsGridOpen
      form.Repaint()
   return

   function MoveRecord(n)
      // If SEEKFIELD has focus and GRID is closed, then
      //  Up or Dn arrow should open the drop-down list.
      //  (see the On Key Label in AF_Seeker onGotFocus() )
      this.ToggleGrid("T") //true
      this.lMatch = true // Found what we're looking for, if arrow keys used
      this.grid.Datalink.next(n)
      // handle Up/Dn arrow'ing too far.. // RA
      if this.grid.Datalink.endofSet // beginning or end
      	this.grid.Datalink.next(-n) // go "back"
      endif
      // updates the SEEKFIELD....
      this.SetValue()
   return

   function SetValue 
      /* Sets value of SEEKFIELD
      	Called by:
      	- MoveRecord  (i.e. up/down arrow keys)
         - Grid_onLeftMouseUp
         - AF_Seeker_Key (after every keystroke)  */
      this.SEEKFIELD.Value := this.grid.Columns["Column1"].DataLink.Value
      // after moving in grid with arrow key, keep focus..
      this.SEEKFIELD.setfocus()
      if this.RefreshAlways
         if type("form.rowset") # "U"
            form.rowset.refreshControls()
         endif
      endif

      if type("this.onChange") # "U"
         this.onChange( (this) )
      endif
   return

   function DisplayEF( cTyped, cFullStr, nPosition )
      /* Called by:  AF_Seeker_oSeek
      	Purpose:  After each keystroke is processed, 
         highlight the correct information.
         Modified (ver 3.1) to handle substrings.
      */
      nPos = nPosition - 1
      cTyped = left(cTyped,nPos)
      cFullStr = left(cFullStr,;
      				MAX(len(trim(cFullStr)),nPos))

      form.Freeze() // in the Baseform

      do case
       case isblank(right(cTyped,1))  // Trailing space
       	this.GRID.Datalink.next(1)   // NB: endofSet doesn't matter
         cNextValue = this.grid.Columns["Column1"].DataLink.Value

			if UPPER(cNextValue) = UPPER(cTyped) // 'abc'='a' is .T.
         	this.lMatch = true
            this.SEEKFIELD.Value := left(cNextValue,;
         			MAX(len(trim(cNextValue)),nPos))
            this.SEEKFIELD.setFocus()
      		KEYB "{End}" + "{Shift+HOME}" ;
          	 + REPLICATE ( "{Shift+RightArrow}", nPos ) CLEAR // nPos+1 for Space
         else
            this.lMatch = false
         endif

		 case UPPER(cFullStr)=UPPER(cTyped)
      	
         this.lMatch = true
         this.SEEKFIELD.Value := cFullStr // already trimmed
         this.SEEKFIELD.setFocus()
      	KEYB "{End}" + "{Shift+HOME}" ;
          + REPLICATE ( "{Shift+RightArrow}", nPos ) CLEAR
      	
       otherwise // new value
      	this.lMatch = false

      endcase

      form.UnFreeze()  // in the BaseForm
      

endclass


CLASS AF_Seeker( oCont ) of ENTRYFIELD( oCont ) custom
protect lNear, lExact, lFound, lRefresh, nMaxPos


   // Set stock properties
   this.FontSize := 9
   this.visible   := false  //RA
   this.top := 1  // slightly below container
   this.selectAll := true
   this.value     := ""
   metric = 6	// Pixels
  
   // Custom properties
   this.rowset    = null    // The seeker's rowset
   this.tabAfter  = 0       // Automatically tab after typing N chars
   this.seek      = this.oSeek

   // Internal properties
   this.lNear     = ( SET( "NEAR"  ) == "ON" )
   this.lExact    = ( SET( "EXACT" ) == "ON" )
   this.lRefresh  = this.form.RefreshAlways
   this.lFound    = true
   this.nMaxPos   = 0       // Max position of cursor in entryfield

   FUNCTION onGotFocus
      this.setfocus()  // definitely needed...
      this.seek     = this.oSeek
      this.seekRowset = this.rowset
      this.lFound  = .t.
      this.nMaxPos = 0

      _app.GridCombo = this.parent
      // arrow handler
      on key label Uparrow _app.GridCombo.MoveRecord(-1)
      on key label Dnarrow _app.GridCombo.MoveRecord(+1)
   return

  FUNCTION key( nChar, nPosition )

    if nChar = 9
      // handle TABbing out
      if this.Parent.IsGridOpen
         this.parent.SetValue()
         this.Parent.ToggleGrid("F")   // Close

         // check to see if we can move the cursor to the next object...
         if type("this.Parent.before.SetFocus") # "U"
            this.Parent.before.SetFocus()
         elseif type("this.Parent.before.first.SetFocus") # "U"
            this.Parent.before.first.SetFocus()
         endif            
         return nChar
      else
         return nChar
      endif                  
    endif

    *if nChar = 32 // spacebar
    *	return nChar
    *endif

    if nextkey() # 0
      // Wait until there are no keystrokes pending, and...
      RETURN ( nChar # 255 )  // ...ignore keystroke if special key
    endif
    if nChar == 255
      if .not. empty( this.value ) .and. ;
          ( this.lFound .or. nPosition <= this.nMaxPos )
        this.lFound = this.seek(nPosition)  // needed in displayEF()
      endif
      if this.tabAfter > 0
        // Stuff tab if position is greater than tabAfter
        nChar = iif( nPosition > this.tabAfter, 9, 0 )
      else
        nChar = 0
      endif
      this.nMaxPos = max( nPosition, ;
          min( this.nMaxPos, len( trim( this.Value ))) )
    else
      if nChar >= 32 and nChar < 255
        // Stuff special key to refire key event
        keyboard chr(255)
      elseif nChar = 13 // enter key pressed
         // If grid is open, set value and close it
         if this.parent.IsGridOpen
         	this.parent.SetValue()
         	this.Parent.ToggleGrid()
         else
         	// grid not open, move on to next control
            KEYBOARD "{TAB}"
         endif
      endif

    endif
 RETURN nChar


  FUNCTION oSeek (nPosition) // nPos needed in displayEF()
    local lRet
    lRet = this.rowset.findKeyNearest( this.normalizedValue() )
    if ( this.rowset.endOfSet )
      this.rowset.next( -1 )
    endif
    // Show partially filled entryfield
    cFullStr = this.parent.GRID.Columns["Column1"].DataLink.Value
    cTyped = this.Value
    this.parent.DisplayEF( cTyped, cFullStr, nPosition )
    RETURN lRet

  FUNCTION normalizedValue (cValue)
    if argCount() < 1 // no parameter passed
    	cValue = this.Value
    endif
    RETURN upper( trim( cValue ))

   FUNCTION onLostFocus    
      /* Bug in action: open the grid, then click away.
      Particularly ugly on a multi-page form...
      Problem is, I need to keep focus on the SEEKFIELD, 
      so that keystrokes will always work to do an 
      incremental seek.  However, I also need the GRID 
      onLeftMouseUp to select a value from the grid with
      the rodent.  Unfortunately, there's no good way to 
      tell whether you're clicking away from the SEEKFIELD
      because you're clicking on the GRID, or whether 
      you're clicking on some other control.  If NextObj
      was working, this bug will disappear....
      Possible workaround:  In the OnGotFocus of all other
      controls, make the grid VISIBLE=false.
      NB: same bug in GridCombo :(
      
      // NextObj is unreliable....
      if type("form.nextobj") =="O"
      	cNextobj = form.nextobj.name
      else
      	cNextobj = "not known"
      endif
      ? this.Name+' onLostFocus(), nextobj = ' +cNextobj
      */
      if form.rowset == null
         if .not. this.lNear
            set near off
         endif
         if this.lExact
            set exact on
         endif
         form.RefreshAlways = this.lRefresh
      endif
      // Swap the datalinked and seeker fields
      this.parent.DATAFIELD.value := this.value
      this.parent.DATAFIELD.visible := true
      this.Visible = false

      // arrow handler
      _app.GridCombo = null
      on key label Uparrow 
      on key label Dnarrow 

      // Execute the NoMatch() function if defined and needed
      if type("this.parent.NoMatch") # "U"
      	if NOT this.parent.lMatch
            this.parent.lMatch := true  // reset
      		this.parent.NoMatch(this.Value)
      	endif
      endif
         			
   return


ENDCLASS

CLASS AF_Data( oCont ) of ENTRYFIELD( oCont ) custom

   // Set stock properties
   this.top := 1  // slightly below container
   this.visible   := true
   metric = 6	// Pixels
   this.FontSize := 9

   FUNCTION onGotFocus
      // When the DataField gets focus, need to
      // 1. Transfer its value into the SeekField
      // 2. Move row pointer in Grid seeker
      // 3. Swap fields, so that SeekField becomes
      //		visible and DataField is hidden
      this.parent.SEEKFIELD.value := this.value
      cSeek = this.parent.SEEKFIELD.normalizedValue(this.Value)
      if NOT this.parent.SEEKFIELD.rowset.findKey( cSeek )
         this.parent.SEEKFIELD.rowset.first()
      endif

      this.parent.SEEKFIELD.visible := true
   	this.visible	:= false
      this.parent.SEEKFIELD.onGotFocus() // setfocus() doesn't
      
   return


   
ENDCLASS

Class AF_Grid(oParent) of Grid(oParent)
   with (this)
      metric = 6	// Pixels
      height = 8 // default, override by dropdownheight
      visible = true
      gridLineWidth = 0
      hasColumnLines = false
      hasRowLines = false
      hasIndicator = false
      hasColumnHeadings = false
      hasVScrollHintText = false
      rowSelect = true
      allowRowSizing = false
      allowColumnSizing = false
      allowColumnMoving = false
      allowEditing = false
      allowAddRows = false
      hScrollBar = 0	// Off
      visible = false
   endwith


   function onLeftMouseUp

      this.parent.SetValue()  // sets value of SEEKFIELD,
                              // sets focus to SEEKFIELD
		this.parent.ToggleGrid() // closes grid
      // Keep the datalinked and seeker fields in sync
      this.parent.DATAFIELD.value :=;
      	 this.parent.SEEKFIELD.value
      this.parent.SEEKFIELD.visible := true
      this.parent.DATAFIELD.visible := false
      
   return


endclass

