/*
   --------------------------------------------------------------------
    Class.....: IndexTags.cc
    Author....: Ken Mayer
    Date......: June 26, 2003
    Notes.....: This class is meant to be usable by a developer in 
                one of a couple ways:

                1) Programmatic -- it allows the developer to return
                                   a multi-column array of information
                                   about the index tags for a table.
                 
                2) If desired, a dialog can be displayed showing
                   a grid of index tags and index expressions, etc.
                   This would allow the user to generate such a list
                   on the fly for their users, for example, that could
                   be used to allow the users to select an index
                   tag, if they way, and change the sequence of 
                   a table.

                *** Warning: this has been tested only against
                  .MDX tags (.DBF files) and some Interbase
                  tables, but it *should* work for all others ...
                  I believe that the database must be closed
                  for InterBase aliases to work ... 

    Updates...: September 9, 2003 -- KJM
                Modified to use Michael Nuwer's new class, BDEIndex.cc
                also in the dUFLP ...

    Usage.....: set procedure to :dUFLP:IndexTags.cc additive
                oIndexes = new IndexTags( <cTableName>[, <cUserid>] )
                aTags = oIndexes.GetTags()
                 // returns an array of tagnames, expressions, and so on
                 // This code actually just uses a tableDef to get
                 // the information and return the array.

                 The array returned has seven columns. 
                 For a .DBF's MDX file:
                    1) IndexName
                    2) Expression
                    3) ForExpression
                    4) Descending
                    5) Unique
                    6) Distinct
                    7) Primary

                 For Non-dBF Indexes:
                    1) IndexName
                    2) Fields
                    3) CaseSensitive
                    4) Descending
                    5) Unique
                    6) <empty> (false)
                    7) Primary (false)

                 Example: If you wanted to find the Primary Key
                          in a dBF table:

                 set procedure to :dUFLP:IndexTags.cc additive
                 oIndexes = new IndexTags( "MyTable.dbf" )
                 aTags = oIndexes.GetTags()
                 for i = 1 to aTags.size/7 // seven columns
                     if aTags[i,7] // if the Primary column is true
                        ? aTags[i,1] // display the indexName
                        exit // out of the loop
                     endif
                 next
 
                    OR

                 cTag = oIndexes.ShowTags()
                 // pops up a form with a grid
                 // This requires the code to create a temporary
                 // table used by the form and the grid to display
                 // the information, but again the information
                 // is really coming from the tableDef object.
    Parameters: cTableName -- can be a plain tablename (test.dbf) or
                              it can include a bde alias 
                              (:myalias:Test.dbf), but it should 
                              include the extension ...
                cUserid    -- This is the userid with
                              the password, in the format
                              of USERID/Password -- the slash
                              is vital. This is optional, but
                              can be used for databases
                              that have logins ...
    Returns...: An array, if using the GetTags method
                    OR
                A character string, being the indexName of the
                  index tag selected by the user if using
                  the ShowTags() method.

       ******* dUFLP NOTE: ******* 
    This uses the following code from the dUFLP, and 
    needs to be deployed with your application if that's
    what you need:
             :dUFLP:BDEAlias.cc
             :dUFLP:BDEIndex.cc
               // which uses:
               :dUFLP:vStruct.h
               :dUFLP:vStruct.prg
    ******* End of dUFLP NOTE ******* 
   --------------------------------------------------------------------
*/

class IndexTags( cTable, cUserid )
   protect MakeArray, SetTable, Error

   // just to make sure we don't display multiple
   // errors ...
   this.Error = false

   if argCount() == 0
      msgbox( "There is no tablename passed to this object --"+;
              "this is required.", "IndexTags Error", 16 )
      return
   endif

   // when instantiated we need to do ensure we have a table
   // and if it's got a BDE alias, we need to work with that ...
   if class::SetTable(cTable, cUserid)
      class::MakeArray()
   else
      if not this.error
         // error:
         msgbox( "The tablename is not valid.", "IndexTags Error", 16 )
      endif
      return
   endif

   function SetTable
      parameters cTable, cUserid
      local bSuccess, oTable, oData
      bSuccess = false
      this.Table = null
      this.Alias = null
      if argCount() => 1 and type("cTable") == "C" 
			if file(cTable)      	
            this.Table = cTable
            bSuccess := true
         elseif cTable.indexOf(":") = 0 and cTable.lastIndexOf(":") > 1
        		// Maybe a database alias -- code borrowed from
            // DBF7File.cc ...:
        		set procedure to :dUFLP:BDEAlias.CC additive
            local b, nAt, cAliasName, cTableName, cDatabasePath

				// Extract alias and tableName 
         	nAt 		  = cTable.lastIndexOf(":")
            cAliasName = cTable.substring(1,nAt)
            cTableName = cTable.Right(cTable.Length-nAt-1)

            // Get database path
		      b = new BDEAlias()
            // Check to see if Alias exists
            if not b.databaseFound( cAliasName )
               msgbox( "BDE Alias does not exist.", "IndexTags Error", 16 )
               this.Error = true
               return bSuccess
            endif

            // Check here for non SQL server databases?
            try
               oData = new Database()
               if ArgCount() == 2 // we have a userid/password string
                  oData.loginString := cUserID
               endif
               oData.databaseName := cAliasName
               oData.active       := true
               oTable = new TableDef()
               oTable.database  := oData
               oTable.TableName := cTableName
               oTable.load()
            catch( Exception E )
               msgbox( "Unexpected error: "+;
                       e.code+" -- "+e.message+" -- Line: "+e.LineNo,;
                       "IndexTags Error", 16 )
               this.Error = true
               return bSuccess
            endtry
            cDatabasePath = b.databaseDir( cAliasname )

            // if here, we've loaded the table ...
            if oTable.tableType == "DBASE" or;
               oTable.tableType == "PARADOX"
               if file( cDatabasePath + "\" + cTableName )
                  this.Table = cDatabasePath + "\" + cTableName
                  this.Alias = cAliasName
                  bSuccess = true
               endif
            else // if we have no error:
               this.Alias = cAliasName
               this.Table = cTableName
               bSuccess = true
            endif
            release object oTable
            close procedure :dUFLP:BDEAlias.CC
         endif
      endif
   return bSuccess
      
   function MakeArray
      // We have a table, let's load the array with the
      // information we need. 
      local oData, oIDX, q, aTagNames, nTags, nIndex, aTagProps
      if not empty( this.Alias )
         oData = new database()
         oData.databaseName := this.Alias
         if ArgCount() == 2 // we have a userid/password string
            oData.loginString := cUserID
         endif
         oData.active       := true
      endif
      // need a query, too, since we're using the :CoreShared:stBDE32
      // class:
      q = new query()
      if not empty( this.alias )
         q.database := oData
      endif
      q.sql := "select * from '"+this.table+"'"
      q.active := true

      // Uses Core Product file and BDE API calls,
      // this allows us to know if a dBF index tag
      // is the Primary Key:
      set proc to :dUFLP:BDEIndex.cc additive
      oIDX = new BDEIndex()

      // use query:
      aTagNames = oIDX.TagList( q ) 
      nIndexes = aTagNames.size
      
      // if the number of tagnames is > 0
      if nIndexes > 0
         // set array size based on tableType:
         this.IndexArray = new Array( 1, 7 )
         // look through indexes array
         for nIndex = 1 to nIndexes
             if nIndex > 1
                this.IndexArray.grow( 1 ) // add a row
             endif
             aTagProps = oIDX.TagProperties( q, nIndex )                
             if q.rowset.tableDriver == "DBASE"
                this.IndexArray[ nIndex, 1 ] := aTagProps[ "IndexName" ]
                this.IndexArray[ nIndex, 2 ] := aTagProps[ "Expression" ]
                this.IndexArray[ nIndex, 3 ] := aTagProps[ "ForExpression" ]
                this.IndexArray[ nIndex, 4 ] := aTagProps[ "Descending" ]
                this.IndexArray[ nIndex, 5 ] := aTagProps[ "Unique" ]
                this.IndexArray[ nIndex, 6 ] := aTagProps[ "Distinct" ]
                this.IndexArray[ nIndex, 7 ] := aTagProps[ "Primary" ]
             else // Non-dBF Index:
                this.IndexArray[ nIndex, 1 ] := aTagProps[ "IndexName" ]
                // loop through number of fields:
                cFields = ""
                for i = 1 to aTagProps[ "NumFieldsInKey"]
                    cFieldNum = "KeyField"+str(i,2,0,"0")
                    cField = aTagProps[ cFieldNum ]
                    cFields += left( cField, at( "(", cField ) - 2 )
                    if i < aTagProps[ "NumFieldsInKey" ]
                       cFields += ";"
                    endif
                next
                this.IndexArray[ nIndex, 2 ] := cFields
                this.IndexArray[ nIndex, 3 ] := aTagProps[ "CaseSensitive" ]
                this.IndexArray[ nIndex, 4 ] := aTagProps[ "Descending" ]
                this.IndexArray[ nIndex, 5 ] := aTagProps[ "Unique" ]
                this.IndexArray[ nIndex, 6 ] := false
                this.IndexArray[ nIndex, 7 ] := aTagProps[ "Primary" ]
             endif
         next
      else
         msgbox( "No index tags associated with this table.", ;
                 "IndexTags Error", 16 )
      endif
      if not empty( this.Alias )
         oData.active := false
         release object oData
      endif
      q.active := false
      release object q
      release object oIDX
      close procedure :dUFLP:BDEIndex.cc
   return

   function GetTags
      // this routine should simply return the array:
   return ( this.IndexArray )

   function ShowTags
      // this routine is more involved. We need to create
      // a temporary BDE Alias pointing to the current folder,
      // a temporary table, and a form, etc. We should return
      // simply the name of the index tag ...
      private oBDE, cPath, oData, oTable, i, cTag

      // Create a temporary BDE Alias:
      set procedure to :dUFLP:BDEAlias.cc additive
      oBDE = new BDEAlias()
      // use the user's temp folder, if the windows environment
      // is empty, then use the current folder:
      cPath = GetEnv("TEMP")
      if empty( cPath )
         cPath = set("DIRE")
      endif
      try
         oBDE.createAlias( "dUFLP_IndexTags", cPath )
      catch( Exception E )
         msgbox( "Unexpected error: "+e.code+" -- "+;
                 e.message, "IndexTags Error", 16 )
         return
      endtry

      // Now we need to create the table:
      open Database dUFLP_IndexTags
      if file( cPath+"\TempIndexTagsTable.dbf" )
         drop table :dUFLP_IndexTags:TempIndexTagsTable
      endif
      create table :dUFLP_IndexTags:TempIndexTagsTable (;
                   IndexName char( 40 ),;
                   IndexExpr char( 100 ) )

      // Create a database object:
      oData = new Database( "dUFLP_IndexTags" )
      oTable = new Query()
      oTable.database := oData
      oTable.sql      := "select * from TempIndexTagsTable"
      oTable.active   := true

      // populate the table with information in the array:
      for i = 1 to this.IndexArray.size / 7 // seven columns
          oTable.rowset.beginAppend()
          oTable.rowset.fields["IndexName"].value := this.IndexArray[ i, 1 ]+;
             iif( this.IndexArray[i,6]==true, " (Primary)", "" )
          if empty( oTable.rowset.fields["IndexName"].value )
             oTable.rowset.fields["IndexName"].value := "Primary Key"
          endif
          oTable.rowset.fields["IndexExpr"].value := this.IndexArray[ i, 2 ]
          oTable.rowset.save()
      next

      // Now to use the form (this is shown after this class,
      // if you wish to modify it, feel free)
      fIndexForm = new IndexTagsForm()
      fIndexForm.text := "Select Index tag for table: "+this.table
      fIndexForm.readModal()
      // when the form is closed:
      cTag = fIndexForm.rowset.fields["IndexName"].value.rightTrim()
      if "(Primary)" $ cTag
         cTag = left( cTag, at( "(Primary)", cTag )-1 )
      endif
      fIndexForm.release()

      // now we're done, a bit of cleanup:
      // drop the table:
      oTable.active := false
      oData.dropTable( "TempIndexTagsTable" )
      // release objects:
      release object oTable
      oData.active := false
      release object oData

      // delete the temporary BDE Alias
      close database dUFLP_IndexTags
      oBDE.deleteAlias( "dUFLP_IndexTags" )
      release object oBDE
      close procedure :dUFLP:BDEAlias.cc

   return (cTag)

endclass

// ====================================================================
// IndexTagsForm -- The form used by the ShowTags method of
// the IndexTags object.
// ====================================================================

** END HEADER -- do not remove this line
//
// Generated on 06/25/2003
//
parameter bModal
local f
f = new IndexTagsForm()
if (bModal)
   f.mdi = false // ensure not MDI
   f.readModal()
else
   f.open()
endif

class IndexTagsForm of FORM
   with (this)
      height = 11.9545
      left = 28.2857
      top = 4.7727
      width = 73.7143
      text = "Select Index Tags for Table"
      autoCenter = true
      mdi = false
      showTaskBarButton = false
      sizeable = false
      maximize = false
      minimize = false
      icon = "resource #12"
   endwith

   this.DUFLP_INDEXTAGS1 = new DATABASE()
   this.DUFLP_INDEXTAGS1.parent = this
   with (this.DUFLP_INDEXTAGS1)
      left = 1
      top = 7.6818
      databaseName = "dUFLP_IndexTags"
      active = true
   endwith

   this.TEMPINDEXTAGSTABLE1 = new QUERY()
   this.TEMPINDEXTAGSTABLE1.parent = this
   with (this.TEMPINDEXTAGSTABLE1)
      left = 10.8571
      top = 8.7273
      database = form.duflp_indextags1
      sql = "select * from tempindextagstable.dbf"
      active = true
   endwith

   this.GRID1 = new GRID(this)
   with (this.GRID1)
      dataLink = form.tempindextagstable1.rowset
      columns["Column1"] = new GRIDCOLUMN(form.GRID1)
      columns["Column1"].dataLink = form.tempindextagstable1.rowset.fields["indexname"]
      columns["Column1"].editorType = 1	// EntryField
      columns["Column1"].width = 20
      columns["Column2"] = new GRIDCOLUMN(form.GRID1)
      columns["Column2"].dataLink = form.tempindextagstable1.rowset.fields["indexexpr"]
      columns["Column2"].editorType = 1	// EntryField
      columns["Column2"].width = 142.8571
      with (columns["Column1"].headingControl)
         value = "Tag Name"
      endwith

      with (columns["Column2"].headingControl)
         value = "Expression"
      endwith

      integralHeight = true
      rowSelect = true
      allowEditing = false
      allowAddRows = false
      height = 10.3636
      left = 0.5714
      top = 1.5
      width = 73
   endwith

   this.TITLETEXT = new TEXT(this)
   with (this.TITLETEXT)
      height = 1.3182
      left = 0
      top = 0
      width = 73.7143
      border = true
      anchor = 2	// Top
      colorNormal = "Navy/BtnFace"
      alignHorizontal = 1	// Center
      fontSize = 14
      text = "Select Index Tag"
   endwith

   this.rowset = this.tempindextagstable1.rowset
endclass


